//---------------------------------------------------//
// $File Name : trigger.v
// $Author    : hsun
// $Company   : Anlogic
// $Version   : v1.5
// $Date      : 2019/07/26
//--------------------------------------------------//

module trigger(
   trig_clk,
   non_bus_din,
   bus_din,
   control,
   overflow_flag,
   stop_flag,
   stop_addr,
   wt_ce,
   wt_en,
   wt_addr
);
   //parameters
   parameter      CTRL_REG_LEN = 93;      //control register length
   parameter      NON_BUS_NODE_NUM = 1;   //non-bus trigger node number
   parameter      BUS_NODE_NUM = 8;       //bus trigger node number
   parameter      BUS_NUM = 1;            //bus number
   parameter      BUS1_WIDTH = 8;         //first bus width
   parameter      BUS2_WIDTH = 8;         //second bus width
   parameter      BUS3_WIDTH = 8;
   parameter      BUS4_WIDTH = 8;
   parameter      BUS5_WIDTH = 8;
   parameter      BUS6_WIDTH = 8;
   parameter      BUS7_WIDTH = 8;
   parameter      BUS8_WIDTH = 8;
   parameter      STOP_LEN = 100;

   localparam     BUS1_START_POS = 0;
   localparam     BUS2_START_POS = BUS1_START_POS + BUS1_WIDTH;
   localparam     BUS3_START_POS = BUS2_START_POS + BUS2_WIDTH;
   localparam     BUS4_START_POS = BUS3_START_POS + BUS3_WIDTH;
   localparam     BUS5_START_POS = BUS4_START_POS + BUS4_WIDTH;
   localparam     BUS6_START_POS = BUS5_START_POS + BUS5_WIDTH;
   localparam     BUS7_START_POS = BUS6_START_POS + BUS6_WIDTH;
   localparam     BUS8_START_POS = BUS7_START_POS + BUS7_WIDTH;
      

   //trigger signals
   input [NON_BUS_NODE_NUM-1:0]    non_bus_din;
   input [BUS_NODE_NUM-1:0]        bus_din;
   //from control registers
   input [CTRL_REG_LEN-1:0]        control;
   //
   input          trig_clk;

   output         overflow_flag;
   output         stop_flag;
   output [15:0]  stop_addr;   
   output         wt_ce;
   output         wt_en;
   output [15:0]  wt_addr;

   wire [21:0]                    main_control;
   wire [4*NON_BUS_NODE_NUM-1:0]  non_bus_control;
   wire [67*BUS_NUM-1:0]          bus_control;
        
   wire           trig_rstn;
   wire           trig_pause;
   wire [1:0]     trig_logic;
// wire [1:0]     trig_position;
   wire [15:0]    trig_length;

   wire           trigger_and_en;
   wire           trigger_or_en;
   wire           trigger_en;

   wire [15:0]    force_acq_len; 

   wire [NON_BUS_NODE_NUM-1:0]   trig_non_bus_en;
   wire [BUS_NUM-1:0]            trig_bus_en;
   
   reg  [NON_BUS_NODE_NUM-1:0]   trig_non_bus_en_ff;
   reg  [BUS_NUM-1:0]            trig_bus_en_ff;

   reg            force_acq_fin;
   reg  [15:0]    force_acq_reg;

   reg            emb_store_en; 
   
   reg            trig_pause_sync0;
   reg            trig_pause_sync1;

//only bus trigger
   generate
   if (NON_BUS_NODE_NUM == 0)
   begin: ONLY_BUS_TRIGGER
                assign bus_control = control[CTRL_REG_LEN-1:CTRL_REG_LEN-67*BUS_NUM];
                assign trigger_and_en = &trig_bus_en_ff;
        assign trigger_or_en  = |trig_bus_en_ff;

        //add 1-stage DFF to optimize timing
                always @(posedge trig_clk or negedge trig_rstn) begin
                        if(~trig_rstn) begin
                                trig_bus_en_ff <= 'd0;
                        end
                        else begin
                                trig_bus_en_ff <= trig_bus_en;  
                        end
                end

   end
   endgenerate

//only non-bus trigger
   generate
   if (BUS_NODE_NUM == 0)
   begin: ONLY_NON_BUS_TRIGGER
        assign non_bus_control = control[4*NON_BUS_NODE_NUM+21:22];
                assign trigger_and_en = &trig_non_bus_en_ff;
        assign trigger_or_en  = |trig_non_bus_en_ff;
        
                //add 1-stage DFF to optimize timing
                always @(posedge trig_clk or negedge trig_rstn) begin
                        if(~trig_rstn) begin
                                trig_non_bus_en_ff <= 'd0;
                        end
                        else begin
                                trig_non_bus_en_ff <= trig_non_bus_en;
                        end
                end

   end
   endgenerate

//both bus trigger and non-bus trigger
   generate
   if (NON_BUS_NODE_NUM != 0 & BUS_NODE_NUM != 0)
   begin: BUS_AND_NON_BUS_TRIGGER
        assign bus_control = control[CTRL_REG_LEN-1:CTRL_REG_LEN-67*BUS_NUM];
        assign non_bus_control = control[4*NON_BUS_NODE_NUM+21:22];
                assign trigger_and_en = (&trig_non_bus_en_ff) & (&trig_bus_en_ff);
                assign trigger_or_en  = (|trig_non_bus_en_ff) | (|trig_bus_en_ff);

                always @(posedge trig_clk or negedge trig_rstn) begin
                        if(~trig_rstn) begin
                                trig_bus_en_ff <= 'd0;
                                trig_non_bus_en_ff <= 'd0;
                        end
                        else begin
                                trig_bus_en_ff <= trig_bus_en;
                                trig_non_bus_en_ff <= trig_non_bus_en;
                        end
                end

   end   
   endgenerate

   assign       trigger_en = (trig_logic == 2'b01) ? trigger_or_en : trigger_and_en;
   
   assign       stop_addr = wt_addr; 

   assign       main_control = control[21:0];

   assign       trig_rstn   = main_control[0];
   assign       trig_pause  = main_control[1];
   assign       trig_logic  = main_control[3:2];
// assign       trig_position = main_control[5:4];       //not use
   assign       trig_length = main_control[21:6];

// fix async-set metastable issue
   always @(posedge trig_clk) begin
                trig_pause_sync0 <= trig_pause;
                trig_pause_sync1 <= trig_pause_sync0;
   end

// generate force_acq_len
   assign       force_acq_len = trig_length - STOP_LEN + 16'd10;

   always @(posedge trig_clk or negedge trig_rstn) begin
                if(~trig_rstn) begin
                        force_acq_reg <= 16'd0;
                        force_acq_fin <= 1'b0;
                end
                else begin
                        if(force_acq_reg < force_acq_len) begin
                                force_acq_reg <= force_acq_reg + 1'b1;
                        end
                        else begin
                                force_acq_fin <= 1'b1;
                        end
                end
   end 

// generate non-bus detector
   genvar i;
   generate
   if (NON_BUS_NODE_NUM != 0)
   begin: HAS_NON_BUS_DETECTOR 
          for(i=0; i<NON_BUS_NODE_NUM; i=i+1)
          begin : NON_BUS_DETECTOR
              detect_non_bus non_bus_detector_inst(
                                        .clk(trig_clk),
                        .rstn(trig_rstn),
                        .node_din(non_bus_din[i]),
                        .force_acq_fin(force_acq_fin),
                        .node_logic(non_bus_control[i*4+3]),
                        .edge_mode(non_bus_control[i*4+2:i*4]),
                        .trigger_node_en(trig_non_bus_en[i])
              );
          end
   end
   endgenerate

// generate bus detector
   genvar j;
   generate
   if (BUS_NODE_NUM != 0)
   begin: HAS_BUS_DETECTOR
        for (j = 0; j < BUS_NUM; j = j+1)
        begin: BUS_DETECTOR 
               if(j == 0)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS1_WIDTH)
                    ) bus_detector_inst1( 
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS1_START_POS+BUS1_WIDTH-1:BUS1_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS1_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 1)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS2_WIDTH)
                    ) bus_detector_inst2(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS2_START_POS+BUS2_WIDTH-1:BUS2_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS2_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 2)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS3_WIDTH)
                    ) bus_detector_inst3(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS3_START_POS+BUS3_WIDTH-1:BUS3_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS3_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 3)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS4_WIDTH)
                    ) bus_detector_inst4(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS4_START_POS+BUS4_WIDTH-1:BUS4_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS4_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 4)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS5_WIDTH)
                    ) bus_detector_inst5(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS5_START_POS+BUS5_WIDTH-1:BUS5_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS5_WIDTH-1:67*j]),
                     .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 5)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS6_WIDTH)
                    ) bus_detector_inst6(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS6_START_POS+BUS6_WIDTH-1:BUS6_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS6_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 6)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS7_WIDTH)
                    ) bus_detector_inst7(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS7_START_POS+BUS7_WIDTH-1:BUS7_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS7_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
               if(j == 7)
               begin
                    detect_bus #(
                      .BUS_WIDTH(BUS8_WIDTH)
                    ) bus_detector_inst8(
                      .clk(trig_clk),
                      .rstn(trig_rstn),
                      .force_acq_fin(force_acq_fin),
                      .bus_din(bus_din[BUS8_START_POS+BUS8_WIDTH-1:BUS8_START_POS]),
                      .bus_logic(bus_control[67*j+66:67*j+64]),
                      .bus_value(bus_control[67*j+BUS8_WIDTH-1:67*j]),
                      .trigger_bus_en(trig_bus_en[j])
                    );
               end
        end
   end
   endgenerate

// generate emb_store_en
   always @(posedge trig_clk or negedge trig_rstn) begin
                if(~trig_rstn) begin
                        emb_store_en <= 1'b0;
                end
                else if(trigger_en) begin
                        emb_store_en <= 1'b1;
                end
   end

// emb_control
   emb_ctrl #(
     .STOP_LEN(STOP_LEN)
   ) emb_ctrl_inst(
     .clk(trig_clk),
     .rstn(trig_rstn),
     .emb_store_en(emb_store_en),
     .pause(trig_pause_sync1),
     .trig_length(trig_length),
     .overflow(overflow_flag),
     .wt_addr(wt_addr),
     .wt_ce(wt_ce),
     .wt_en(wt_en),
     .stop(stop_flag)
   );
   

endmodule
