///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2011-2021 Anlogic, Inc.
// All Right Reserved.
///////////////////////////////////////////////////////////////////////////////
//
//
//   Vendor : Anlogic
//   Version : 0.1
//   Description : Anlogic Functional Simulation Library Component
//            bram
//   Filename : ef2_logic_bram.v
//   Timestamp : Mon Nov  4 14:20:41 CST 2013
//
///////////////////////////////////////////////////////////////////////////////
//
// Revision:
//    11/04/13 - Initial version.
//
///////////////////////////////////////////////////////////////////////////////
`timescale 1 ns / 1 ps
module AL_LOGIC_BRAM (
  dia,
  addra, cea, ocea, clka, wea, rsta, bea,
  dib,
  addrb, ceb, oceb, clkb, web, rstb, beb,
  doa,
  dob);
  // 1kx9
  parameter DATA_WIDTH_A = 9;
  parameter DATA_WIDTH_B = DATA_WIDTH_A;
  parameter ADDR_WIDTH_A = 10;
  parameter ADDR_WIDTH_B = ADDR_WIDTH_A;
  parameter DATA_DEPTH_A = 2 ** ADDR_WIDTH_A;
  parameter DATA_DEPTH_B = 2 ** ADDR_WIDTH_B;
  parameter BYTE_ENABLE = 0;            // only interger {0, 8, 9} are valid
  parameter BYTE_A = BYTE_ENABLE == 0 ? 1 : DATA_WIDTH_A / BYTE_ENABLE;
  parameter BYTE_B = BYTE_ENABLE == 0 ? 1 : DATA_WIDTH_B / BYTE_ENABLE;
  output  [DATA_WIDTH_A-1:0] doa;
  output  [DATA_WIDTH_B-1:0] dob;
  input   [DATA_WIDTH_A-1:0] dia; //DEFAULT := (others => '0')
  input   [DATA_WIDTH_B-1:0] dib; //DEFAULT := (others => '0')
  input   cea, ocea, clka, wea, rsta; //DEFAULT := '0'
  input   ceb, oceb, clkb, web, rstb; //DEFAULT := '0'
  input [BYTE_A - 1 : 0] bea; //DEFAULT := (others => '0')
  input [BYTE_B - 1 : 0] beb; //DEFAULT := (others => '0')
  input   [ADDR_WIDTH_A-1:0] addra; //DEFAULT := (others => '0')
  input   [ADDR_WIDTH_B-1:0] addrb; //DEFAULT := (others => '0')
  parameter MODE = "DP";                 // DP,SP,PDPW,FIFO
  parameter REGMODE_A = "NOREG";         // "NOREG", "OUTREG"
  parameter REGMODE_B = "NOREG";         // "NOREG", "OUTREG"
  parameter WRITEMODE_A = "NORMAL";      // "NORMAL", "READBEFOREWRITE", "WRITETHROUGH"
  parameter WRITEMODE_B = "NORMAL";      // "NORMAL", "READBEFOREWRITE", "WRITETHROUGH"
  parameter RESETMODE = "SYNC";          // "SYNC", "ASYNC"
  parameter DEBUGGABLE = "NO";           // "YES", "NO"
  parameter PACKABLE = "NO";             // "YES", "NO"
  parameter FORCE_KEEP = "OFF";          // "ON", "OFF"
  parameter INIT_FILE = "NONE";
  parameter FILL_ALL = "NONE";          // "NONE" or some binary string like "0101", not wider than the wider port
  parameter IMPLEMENT = "9K";            // 9K | 9K(FAST) | 32K (all capitalized)
  localparam RD_DELAY = 0.2;
  localparam DFF_ODELAY = 0.1;
  localparam SIM_INIT = 0;
tri0 rsta;
tri0 rstb; 
// TO BE FILLED
//****************************************
//         Port Definitions
//****************************************
  // clk : Port clk input, posedge synchronize read and write opeartion.
  // ce  : Port clk enable
  // we  : Port write/read operation 1 write, 0 read
  // be  : Port byte write enable, only DATA_WIDTH >=16 will use it  
  // rst : Port reset input to reset the doa output register when REGMODE=OUTREG 
  // oce : Port output register clk enable input, only be used when REGMODE=OUTREG , oce is higher priority than rst
  // di  : Port write data input bus
  // addr: Port write/read address input bus
  // do  : Port output data bus
  //
  // ***** write/read operation control logic********************
  // ce we 
  //  0   x  idle/no change
  //  1   1  write operation
  //  1   0  read  operation   
//****************************************
//     some Block Parameter Definitions
//****************************************
  // MODE :
  //   value "DP" : Dual Port RAM, Port-A write/read, Port-B write/read
  //   value "SP" : Simple Port RAM, only Port-A write/read, user can't use Port-B
  //   value "PDPW" : Port-A write, Port-B read, user must connect wea=1 , web=0, float doa/ocea/rsta and dib
  // RESETMODE :
  //   value "SYNC" : sync reset output register of doa and dob
  //   value "ASYNC" : async reset output register of doa and dob
  // BYTE_ENABLE : 
  //       value must be 0 or 8 or 9
  //       0: no byte enable function , bea is not used
  //       8: means every 8 bit vs one bea bit, DATA_WIDTH must be 8*N , N is >=2 integer
  //       9: means every 9 bit vs one bea bit, DATA_WIDTH must be 9*N , N is >=2 integer
  // others :
  //   reference Datasheet or PHY model

  localparam DATA_WIDTH_MEM = (DATA_WIDTH_A < DATA_WIDTH_B) ? DATA_WIDTH_A : DATA_WIDTH_B; // select min data width 
  localparam DEPTH_MEM_A = DATA_DEPTH_A * DATA_WIDTH_A / DATA_WIDTH_MEM ;
  localparam DEPTH_MEM_B = DATA_DEPTH_B * DATA_WIDTH_B / DATA_WIDTH_MEM ;            
  localparam DEPTH_MEM = (DEPTH_MEM_A > DEPTH_MEM_B) ?  DEPTH_MEM_A : DEPTH_MEM_B; // select max data depth

  
  localparam RATIO_A = DATA_WIDTH_A / DATA_WIDTH_MEM ;
  localparam RATIO_B = DATA_WIDTH_B / DATA_WIDTH_MEM ;
  localparam BYTE_WIDTH = (BYTE_ENABLE == 9) ? 9 : 8;  


  wire [BYTE_A - 1 : 0] byte_ena;
  wire [BYTE_B - 1 : 0] byte_enb;
  reg [DATA_WIDTH_A-1:0] mem_q_a,mem_q_a_tmp;
  reg [DATA_WIDTH_B-1:0] mem_q_b,mem_q_b_tmp;
  reg [DATA_WIDTH_MEM-1:0] memory [0:DEPTH_MEM-1];


  

 
  
  generate
  genvar i;    
  if( BYTE_A == 1 ) begin : gen_nobe_a
    assign byte_ena[0] = 1'b1;
  end          
  else begin   
    for ( i=0; i<BYTE_A; i=i+1 ) begin : gen_byte_ena
      assign byte_ena[i] = bea[i];   
    end        
  end          
  
  if( BYTE_B == 1 ) begin
    assign byte_enb[0] = 1'b1;
  end          
  else begin   
    for ( i=0; i<BYTE_B; i=i+1 ) begin : gen_byte_enb
      assign byte_enb[i] = beb[i];   
    end        
  end          
  endgenerate  
               
               
  //********************
  // Memory Port-A write
  //********************
  task MEM_WRITE_A
    ( input  reg [ADDR_WIDTH_A-1:0]   addr,
      input  reg [BYTE_A-1:0] byte_en,
      input  reg [DATA_WIDTH_A-1:0] data
	);     
    reg [DATA_WIDTH_A-1:0] old_data;
    integer i; 
    begin      
      // read old data first
      if (RATIO_A == 1) begin            
        old_data = memory[addr];
      end else begin
        for (i = 0; i < RATIO_A; i = i + 1) begin
          old_data[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM]= memory[addr*RATIO_A + i];
        end    
      end      
      if(BYTE_A==1) old_data = data;
      else begin 
      // update data based on byte_en
        for (i = 0; i < BYTE_A; i = i + 1) begin
          if (byte_en[i]) begin
            old_data[BYTE_WIDTH*i+:BYTE_WIDTH] = data[BYTE_WIDTH*i+:BYTE_WIDTH];
          end  
        end    
      end         
      // Write data to memory
      if ( RATIO_A == 1) begin
        memory[addr] = old_data;
      end else begin
        for (i = 0; i < RATIO_A; i = i + 1) begin
          memory[addr*RATIO_A + i] = old_data[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM];
        end    
      end      
    end        
  endtask      
               
  //********************
  // Memory Port-B write
  //********************
  task MEM_WRITE_B
    ( input  reg [ADDR_WIDTH_B-1:0] addr,
      input  reg [BYTE_B-1:0] byte_en,
      input  reg [DATA_WIDTH_B-1:0] data
	);     
    reg [DATA_WIDTH_B-1:0] old_data;
    integer i; 
    begin      
      // read old data first
      if (RATIO_B == 1) begin            
        old_data = memory[addr];
      end else begin
        for (i = 0; i < RATIO_B; i = i + 1) begin
          old_data[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM]= memory[addr*RATIO_B + i];
        end    
      end 
      if(BYTE_B==1) old_data = data;
      else begin     
      // update data based on byte_en
        for (i = 0; i < BYTE_B; i = i + 1) begin
          if (byte_en[i]) begin
            old_data[BYTE_WIDTH*i+:BYTE_WIDTH] = data[BYTE_WIDTH*i+:BYTE_WIDTH];
          end  
        end    
      end         
    // Write data to memory
      if ( RATIO_B == 1) begin
        memory[addr] = old_data;
      end else begin
        for (i = 0; i < RATIO_B; i = i + 1) begin
          memory[addr*RATIO_B + i] = old_data[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM];
        end    
      end      
    end        
  endtask      
               
  //********************
  // Memory Port-A read
  //********************
  task MEM_READ_A
    ( input reg [ADDR_WIDTH_A-1:0] addr
      );       
    integer i; 
    begin      
     if (RATIO_A==1) begin
       mem_q_a <= #RD_DELAY memory[addr];
     end else begin
       for (i = 0; i < RATIO_A; i = i + 1) begin
         mem_q_a[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM] <= #RD_DELAY memory[addr*RATIO_A + i];
       end     
     end       
    end        
  endtask      

  task MEM_READ_A_BASEBE
    ( input reg [ADDR_WIDTH_A-1:0] addr,
      input reg [BYTE_A-1:0] be
      );       
    integer i; 
    begin      
     if (RATIO_A==1) begin
       mem_q_a_tmp <= #RD_DELAY memory[addr];
     end else begin
       for (i = 0; i < RATIO_A; i = i + 1) begin
         mem_q_a_tmp[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM] <= #RD_DELAY memory[addr*RATIO_A + i];
       end     
     end
     for (i = 0; i < BYTE_A; i = i + 1) begin
       if (!be[i]) begin
         mem_q_a[BYTE_WIDTH*i+:BYTE_WIDTH] = mem_q_a_tmp[BYTE_WIDTH*i+:BYTE_WIDTH];
       end  
     end      
       
    end        
  endtask      

  task MEM_READ_B_BASEBE
    ( input reg [ADDR_WIDTH_B-1:0] addr,
      input reg [BYTE_B-1:0] be
      );       
    integer i; 
    begin      
     if (RATIO_B==1) begin
       mem_q_b_tmp <= #RD_DELAY memory[addr];
     end else begin
       for (i = 0; i < RATIO_B; i = i + 1) begin
         mem_q_b_tmp[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM] <= #RD_DELAY memory[addr*RATIO_B + i];
       end     
     end
     for (i = 0; i < BYTE_B; i = i + 1) begin
       if (!be[i]) begin
         mem_q_b[BYTE_WIDTH*i+:BYTE_WIDTH] = mem_q_b_tmp[BYTE_WIDTH*i+:BYTE_WIDTH];
       end  
     end      
       
    end        
  endtask     
               
  //*********************
  // Memory Port-B read
  //*********************
  task MEM_READ_B
    ( input reg [ADDR_WIDTH_B-1:0] addr
       );      
    integer i; 
    begin      
     if (RATIO_B==1) begin
       mem_q_b <= #RD_DELAY memory[addr];
     end else begin
       for (i = 0; i < RATIO_B; i = i + 1) begin
         mem_q_b[DATA_WIDTH_MEM*i+:DATA_WIDTH_MEM] <= #RD_DELAY memory[addr*RATIO_B + i];
       end     
     end       
    end        
  endtask      
               
               
               
  //**************************************
  // Memory Port-A & Port-B Read/Write function generations
  //**************************************
  wire cea_gate,ceb_gate;
  wire wea_gate,web_gate;
  wire rea_gate,reb_gate;
  generate     
    if( IMPLEMENT == "9K" || IMPLEMENT == "9K(FAST)" ) begin : gen_ce
      assign cea_gate = !rsta & cea;
      assign ceb_gate = !rstb & ceb;
    end else begin
      assign cea_gate = cea;
      assign ceb_gate = ceb;
    end        
  endgenerate

  generate     
    if( BYTE_ENABLE == 8 || BYTE_ENABLE == 9 ) begin : gen_we
      assign wea_gate = | bea;
      assign web_gate = | beb;
      assign rea_gate = ~(&bea);
      assign reb_gate = ~(&beb);
    end else begin
      assign wea_gate = wea;
      assign web_gate = web;
      assign rea_gate = ~wea;
      assign reb_gate = ~web;
    end        
  endgenerate
  
  
  generate     
    if( WRITEMODE_A == "WRITETHROUGH" ) begin : gen_wtth_a 
      always @(posedge clka) begin
        //only Read A
        if ( cea_gate==1 && rea_gate==1 && wea_gate==0 ) MEM_READ_A(addra);
        //Write A
        else if ( cea_gate==1 && wea_gate==1 ) begin
          MEM_WRITE_A(addra, byte_ena, dia);
          MEM_READ_A(addra);
        end      
      end      
    end        
    else if( WRITEMODE_A == "READBEFOREWRITE" ) begin : gen_rdbfwt_a
      always @(posedge clka) begin
        //only Read A
        if ( cea_gate==1 && rea_gate==1 && wea_gate==0 ) MEM_READ_A(addra);
        //Write A after 
        else if ( cea_gate==1 && wea_gate==1 ) begin
          MEM_READ_A(addra);
          MEM_WRITE_A(addra, byte_ena, dia);
        end    
      end      
    end        
    else if( WRITEMODE_A == "NORMAL" ) begin : gen_wtonly_a 
      always @(posedge clka) begin
        //Read A 
        if ( cea_gate==1 &&  rea_gate==1 && wea_gate==0 ) MEM_READ_A(addra);
        //Write A 
        else if ( cea_gate==1 && wea_gate==1 && rea_gate==0 ) begin
          MEM_WRITE_A(addra, byte_ena, dia);
        end
        else if ( cea_gate==1 && wea_gate==1 && rea_gate==1 ) begin
          MEM_READ_A_BASEBE(addra, byte_ena);
          MEM_WRITE_A(addra, byte_ena, dia);
        end
      end      
    end        
  endgenerate  
               
  generate     
    if( WRITEMODE_B == "WRITETHROUGH" ) begin : gen_wtth_b 
      always @(posedge clkb) begin
        //only Read B
        if ( ceb_gate==1 && reb_gate==1 && web_gate==0 ) MEM_READ_B(addrb);
        //Write B
        if ( ceb_gate==1 && web_gate==1 ) begin
          MEM_WRITE_B(addrb, byte_enb, dib);
          MEM_READ_B(addrb);
        end      
      end      
    end        
    else if( WRITEMODE_B == "READBEFOREWRITE" ) begin : gen_rdbfwt_b
      always @(posedge clkb) begin
        //only Read B
        if ( ceb_gate==1 && reb_gate==1 && web_gate==0 ) MEM_READ_B(addrb);
        //Write B after 
        if ( ceb_gate==1 && web_gate==1 ) begin
          MEM_READ_B(addrb);
          MEM_WRITE_B(addrb, byte_enb, dib);
        end    
      end      
    end        
    else if( WRITEMODE_B == "NORMAL" ) begin : gen_wtonly_b 
      always @(posedge clkb) begin
        //Read B 
        if ( ceb_gate==1 &&  reb_gate==1 && web_gate==0 ) MEM_READ_B(addrb);
        //Write B
        else if ( ceb_gate==1 && web_gate==1 && reb_gate==0  ) begin
          MEM_WRITE_B(addrb, byte_enb, dib);
        end 
        else if ( ceb_gate==1 && web_gate==1 && reb_gate==1  ) begin
          MEM_READ_B_BASEBE(addrb, byte_enb);
          MEM_WRITE_B(addrb, byte_enb, dib);

        end
      end      
    end        
  endgenerate  
               
  //**************************************
  // Memory Port-A & Port-B output register function generations
  //**************************************

  generate     
    if( (REGMODE_A == "OUTREG") && (RESETMODE == "SYNC") ) begin : gen_outreg_sync_a
      reg [DATA_WIDTH_A-1:0] doa_q=0; 
      always @(posedge clka) begin
        if( ocea ) begin
          if( rsta ) doa_q=0;
          else doa_q = #DFF_ODELAY mem_q_a;
        end    
      end
      assign doa = doa_q;      
    end        
    else if( (REGMODE_A == "OUTREG") && (RESETMODE == "ASYNC") ) begin : gen_outreg_async_a
      reg [DATA_WIDTH_A-1:0] doa_q=0;
      always @(posedge clka or posedge rsta) begin
        if( rsta ) doa_q=0;
        else if( ocea ) doa_q = #DFF_ODELAY mem_q_a;
      end
      assign doa = doa_q;      
    end        
    else begin : gen_latch_a
      assign doa = mem_q_a;
    end        
  endgenerate  

  
  generate     
    if( (REGMODE_B == "OUTREG") && (RESETMODE == "SYNC") ) begin : gen_outreg_sync_b
      reg [DATA_WIDTH_B-1:0] dob_q=0; 
      always @(posedge clkb) begin
        if( oceb ) begin
          if( rstb ) dob_q<=0;
          else dob_q <= #DFF_ODELAY mem_q_b;
        end    
      end
      assign dob = dob_q;             
    end        
    else if( (REGMODE_B == "OUTREG") && (RESETMODE == "ASYNC") ) begin : gen_outreg_async_b
      reg [DATA_WIDTH_A-1:0] dob_q=0;
      always @(posedge clkb or posedge rstb) begin
        if( rstb ) dob_q=0;
        else if( oceb ) dob_q = #DFF_ODELAY mem_q_b;
      end
      assign dob = dob_q;            
    end        
    else begin : gen_latch_b
      assign dob = mem_q_b;
    end        
  endgenerate                 
  //**************************************
  // 9K latch output auto power up reset to zero
  // 32K latch output X 
  //**************************************
  reg check=0;
  
  task init_memory;
    integer i;
    if( (SIM_INIT==0) ) begin
      $display("incase\n");
      for(i=0;i<DEPTH_MEM;i=i+1) begin
        memory[i] = 0;
      end
    end else begin
      $display("error\n");
    end
  endtask
  generate     
    if( IMPLEMENT == "9K" ) begin : gen_initial_9klatch
      integer i;
      initial begin
        mem_q_a = 0;
        mem_q_b = 0;
      end
    end        
  endgenerate  
  
  initial begin
    init_memory;
  end  
  
endmodule      
