///////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2011-2021 Anlogic, Inc.
// All Right Reserved.
///////////////////////////////////////////////////////////////////////////////
//
//   Vendor : Anlogic
//   Version : 0.1
//   Description : Anlogic Functional Simulation Library Component
//            global clk,  io clock, clk div
//   Filename : ef2_phy_clk.v
//   Timestamp : Thu Apr  24 09:30:30 CST 2014
//
///////////////////////////////////////////////////////////////////////////////
//
// Revision:
//    4/24/14 - Initial version.
//
//    8/8/2018 - Wenyan
//               Remove EF2_PHY_DCE (Dynamic Clock Enable)
//               Chen doesn't support this model
//
//
///////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ps

module EF2_PHY_GCLK (clki, clko);
    input clki;
    output clko;
    
    buf (clko,clki);
   
 specify                                      
        (clki  => clko) = (0:0:0, 0:0:0);
 endspecify 
endmodule

module EF2_PHY_IOCLK (clki, stop, clko);
    input clki;
    input stop;
    output clko;

    wire clko_tmp;
    buf (clko,clko_tmp);
    parameter  STOPCLK = "DISABLE"; // ENABLE,DISABLE

    reg ck_en1,ck_en2;
    reg stop_en;

    initial begin
        ck_en1 = 1'b0;
        ck_en2 = 1'b0;
    end

    initial begin
        case (STOPCLK)
            "DISABLE" : stop_en = 1'b0;
            "ENABLE"  : stop_en = 1'b1;
            default   : begin
                            $display("Parameter Warning : The attribute STOPCLK on ECLK instance %m is set to %s.  Legal values for this attribute are ENABLE or DISABLE.", STOPCLK);
                            //$finish;
                        end
        endcase
    end

    assign clksel = ~(stop & stop_en);

    always @(negedge clki)
        if(clksel==1'b0) begin
            ck_en1 <= 1'b0;
            ck_en2 <= ck_en1;
        end
        else begin
            ck_en1 <= 1'b1;
            ck_en2 <= ck_en1;
        end

    assign clko_tmp = stop_en ? (clki & ck_en2) : clki;

wire done_gwe = glbl.done_gwe;
 specify                                      
        (clki  => clko) = (0:0:0, 0:0:0);
 	$setuphold (posedge clki, posedge stop &&& (done_gwe!=0)	, 0:0:0, 0:0:0);
 	$setuphold (negedge clki, posedge stop &&& (done_gwe!=0)	, 0:0:0, 0:0:0); 
 	$setuphold (posedge clki, negedge stop &&& (done_gwe!=0)	, 0:0:0, 0:0:0);
 	$setuphold (negedge clki, negedge stop &&& (done_gwe!=0)	, 0:0:0, 0:0:0); 
 		
 endspecify
endmodule
 
module EF2_PHY_CLKDIV (clki, rst, rls, clkdiv1, clkdivx);

    output clkdiv1, clkdivx;
    input clki;
    input rst;
    input rls;

    tri1 gsrn = glbl.gsrn;
    //tri1 gsrn;

    parameter GSR = "DISABLE";  //  DISABLE/ENABLE 
    parameter DIV = 2;          //  2/4

//check parameter
    initial begin
        if (DIV != 2 && DIV != 4) begin
        $display("attribute syntax error : the attribute DIV on clkdiv instance %m is set to %d. legal values for this attribute are 2 or 4.", DIV);
        $finish;
        end
    end

    reg sr_reg1, sr_reg2, SRN, rst_d1, set;
    wire clkdiv1_val, clkdiv1_buf, sr;
    reg [2:0] count_dn;
    reg clki_buf;

    buf inst_buf0 (clkdiv1, clkdiv1_buf);
    wire clkdivx_tmp;
    buf (clkdivx,clkdivx_tmp);
    assign clkdivx_tmp = (DIV == 2) ? count_dn[0] : count_dn[1];

    initial 
    begin
        count_dn = 2'b00;
    end

    always @ (gsrn) begin
        if (GSR == "ENABLE") begin
            SRN = gsrn;
        end
        else if (GSR == "DISABLE")
            SRN = 1'b1;
    end

    not (SR1, SRN);
    or (sr, rst, SR1);

    always @ (posedge clki)
    begin
        sr_reg1 <= sr;
    end

    always @ (posedge clki)
    begin
        sr_reg2 <= sr_reg1;
    end

    always @ (clki)
    begin
        clki_buf <= clki;
    end

    and (clkdiv1_val, clki_buf, ~sr_reg1);
    and (clkdiv1_buf, clkdiv1_val, ~sr_reg2);

    always @ (posedge clki or posedge sr)
    begin
        if (sr == 1'b1)
            rst_d1 <= 1'b0;
        else
            rst_d1 <= 1'b1;
    end

    and (setn, rst_d1, rls);
    always @ (posedge clki or negedge setn)
    begin
        if (setn == 1'b0)
            count_dn <= 2'b00;
        else
            count_dn <= count_dn - 1;
    end
wire done_gwe = glbl.done_gwe;    
 specify                                      
        (clki  => clkdiv1) = (0:0:0, 0:0:0);
        (clki  => clkdivx) = (0:0:0, 0:0:0);        
 	$setuphold (posedge clki, posedge rls &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);
 	$setuphold (negedge clki, posedge rls &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);
 	$setuphold (posedge clki, posedge rst &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);
 	$setuphold (negedge clki, posedge rst &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0); 

 	$setuphold (posedge clki, negedge rls &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);
 	$setuphold (negedge clki, negedge rls &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);
 	$setuphold (posedge clki, negedge rst &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);
 	$setuphold (negedge clki, negedge rst &&& (done_gwe!=1'b0)	, 0:0:0, 0:0:0);  		 	
 endspecify  
endmodule

